
from syned.storage_ring.light_source import LightSource
from syned.storage_ring.electron_beam import ElectronBeam
from syned.storage_ring.magnetic_structures.undulator import Undulator
from syned.beamline.element_coordinates import ElementCoordinates
from syned.beamline.beamline_element import BeamlineElement
from syned.beamline.optical_elements.absorbers.slit import Slit
from syned.beamline.optical_elements.mirrors.mirror import Mirror
from syned.beamline.shape import Rectangle, Ellipsoid

from wofry.beamline.decorators import WOLightSourceDecorator, WOOpticalElementDecorator
from wofry.propagator.decorators import WavefrontDecorator
from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D, WavefrontDimension
from wofry.propagator.propagator import PropagationManager, AbstractPropagator, PropagationParameters, PropagationElements

# ----------------------------------------------
# ENGINE 1 (es. SRW)

class WavefrontEngine1(GenericWavefront2D, WavefrontDecorator):

    def __init__(self):
        self._test_parameter = 0

    def set_test_parameter(self, value):
        self._test_parameter = value

    def get_test_parameter(self):
        return self._test_parameter

    def toGenericWavefront(self):
        return self

    def fromGenericWavefront(self, wavefront):
        self.set_test_parameter(0)

class UndulatorEngine1(LightSource, WOLightSourceDecorator):

    def __init__(self):
        super().__init__(name="UndulatorTest",
                         electron_beam=ElectronBeam.initialize_as_pencil_beam(energy_in_GeV=2.0,
                                                                              energy_spread=0.0008,
                                                                              current=0.3),
                         magnetic_structure=Undulator.initialize_as_vertical_undulator(K=1.0223,
                                                                                       period_length=0.01,
                                                                                       periods_number=22))

    def get_wavefront(self):
        wavefront = WavefrontEngine1()
        wavefront.set_test_parameter(100)

        return wavefront


class SlitEngine1(Slit, WOOpticalElementDecorator):

    def __init__(self):
        Slit.__init__(self,
                      name="SlitTest",
                      boundary_shape=Rectangle(x_left=-0.01, x_right=0.01, y_bottom=-0.01, y_top=0.01))
        WOOpticalElementDecorator.__init__(self)

    def applyOpticalElement(self, wavefront):
        wavefront.set_test_parameter(10)

        return wavefront

class PropagatorEngine1(AbstractPropagator):

    HANDLER_NAME = "TEST1"

    def get_dimension(self):
        return WavefrontDimension.TWO

    def get_handler_name(self):
        return self.HANDLER_NAME

    def do_propagation(self, parameters=PropagationParameters()):
        wavefront = parameters.get_wavefront()

        if not isinstance(wavefront, WavefrontEngine1): raise Exception("wrong wavefront! it is not WavefrontTest")

        for element in parameters.get_PropagationElements().get_propagation_elements():
            wavefront = element.get_optical_element().applyOpticalElement(wavefront)

        return wavefront

# -------------------------------------------------------
# ENGINE 2

class WavefrontEngine2(GenericWavefront1D, WavefrontDecorator):

    def __init__(self):
        self._test_parameter_1 = 10
        self._test_parameter_2 = 20

    def set_test_parameter_1(self, value):
        self._test_parameter_1 = value

    def get_test_parameter_1(self):
        return self._test_parameter_1

    def set_test_parameter_2(self, value):
        self._test_parameter_2 = value

    def get_test_parameter_2(self):
        return self._test_parameter_2

    def toGenericWavefront(self):
        return self

    def fromGenericWavefront(self, wavefront):
        self.set_test_parameter_1(10)
        self.set_test_parameter_2(20)

class MirrorEngine2(Mirror, WOOpticalElementDecorator):

    def __init__(self):
        Mirror.__init__(self,
                        name="MirrorEngine2",
                        boundary_shape=Rectangle(x_left=-10.0, x_right=10.0, y_bottom=-10.0, y_top=10.0),
                        surface_shape=Ellipsoid())
        WOOpticalElementDecorator.__init__(self)

    def applyOpticalElement(self, wavefront):
        wavefront.set_test_parameter_1(30)
        wavefront.set_test_parameter_2(40)

        return wavefront

class PropagatorEngine2(AbstractPropagator):

    HANDLER_NAME = "TEST2"

    def get_dimension(self):
        return WavefrontDimension.ONE

    def get_handler_name(self):
        return self.HANDLER_NAME

    def do_propagation(self, parameters=PropagationParameters()):
        wavefront = parameters.get_wavefront()

        if not isinstance(wavefront, WavefrontEngine2): raise Exception("wrong wavefront!  it is not WavefrontEngine2")

        for element in parameters.get_PropagationElements().get_propagation_elements():
            wavefront = element.get_optical_element().applyOpticalElement(wavefront)

        return wavefront

if __name__=="__main__":

    # -------------- INITIALIZATION OF CHAIN OF RESPONSIBILITY
    #
    # MUST BE DONE ONCE

    propagator = PropagationManager.Instance()
    propagator.add_propagator(PropagatorEngine1())
    propagator.add_propagator(PropagatorEngine2())

    #---------------------------

    light_source = UndulatorEngine1()
    propagation_elements = PropagationElements()
    propagation_elements.add_beamline_element(BeamlineElement(coordinates=ElementCoordinates(), 
                                                              optical_element=SlitEngine1()))

    initial_wavefront = light_source.get_wavefront()

    print("TEST1 initial WF:           test parameter = ", initial_wavefront.get_test_parameter())

    wavefront1 = propagator.do_propagation(PropagationParameters(wavefront=light_source.get_wavefront(),
                                                                propagation_elements=propagation_elements),
                                          handler_name="TEST1")

    print("TEST1 : propagate from source to Slit")

    print("TEST1 after WF propagation: test parameter = ", wavefront1.get_test_parameter())

    propagation_elements2 = PropagationElements()
    propagation_elements2.add_beamline_element(BeamlineElement(coordinates=ElementCoordinates(), optical_element=MirrorEngine2()))

    print("switch to TEST2 from TEST1 - > propagate from slit to mirror")

    initial_wavefront2 = WavefrontEngine2()
    initial_wavefront2.fromGenericWavefront(wavefront1.toGenericWavefront().get_Wavefront1D_from_histogram(axis=1))

    print("TEST2 initial WF:           test parameters = ", initial_wavefront2.get_test_parameter_1(), ", ", initial_wavefront2.get_test_parameter_2())

    final_wavefront = propagator.do_propagation(PropagationParameters(wavefront=initial_wavefront2,
                                                                      propagation_elements=propagation_elements2),
                                                handler_name="TEST2")

    print("TEST2 after WF propagation: test parameters = ", final_wavefront.get_test_parameter_1(), ", ", final_wavefront.get_test_parameter_2())




import numpy

from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
from wofry.propagator.propagator import Propagator1D

class FresnelZoom1D(Propagator1D):

    HANDLER_NAME = "FRESNEL_ZOOM_1D"

    def get_handler_name(self):
        return self.HANDLER_NAME

    def do_specific_progation_after(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    def do_specific_progation_before(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)


    def do_specific_progation(self, wavefront1, propagation_distance, parameters):


        try:
            m = parameters.get_additional_parameter("magnification_x")
        except:
            m = 1.0

        wavefront = wavefront1.duplicate()
        shape = wavefront.size()
        delta = wavefront.delta()
        wavenumber = wavefront.get_wavenumber()
        wavelength = wavefront.get_wavelength()

        fft_scale = numpy.fft.fftfreq(shape)/delta

        x = wavefront.get_abscissas()

        x_rescaling = wavefront.get_abscissas() * m

        r1sq = x ** 2 * (1 - m)
        r2sq = x_rescaling ** 2 * ((m - 1) / m)
        fsq = (fft_scale ** 2 / m)

        Q1 = wavenumber / 2 / propagation_distance * r1sq
        Q2 = numpy.exp(-1.0j * numpy.pi * wavelength * propagation_distance * fsq)
        Q3 = numpy.exp(1.0j * wavenumber / 2 / propagation_distance * r2sq)

        wavefront.add_phase_shift(Q1)

        fft = numpy.fft.fft(wavefront.get_complex_amplitude())
        ifft = numpy.fft.ifft(fft * Q2) * Q3 / numpy.sqrt(m)

        wf_propagated = GenericWavefront1D.initialize_wavefront_from_arrays(x_rescaling,
                                                                            ifft,
                                                                            wavelength=wavelength)

        return wf_propagated





from wofry.propagator.propagator import PropagationManager, PropagationElements, PropagationParameters
from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
from wofry.propagator.propagators1D.fresnel import Fresnel1D
from wofry.propagator.propagators1D.fresnel_convolution import FresnelConvolution1D
from wofry.propagator.propagators1D.fraunhofer import Fraunhofer1D
from wofry.propagator.propagators1D.integral import Integral1D
from wofry.propagator.propagators1D.fresnel_zoom import FresnelZoom1D
from wofry.propagator.propagators1D.fresnel_zoom_scaling_theorem import FresnelZoomScaling1D

from wofry.beamline.optical_elements.ideal_elements.screen import WOScreen
from wofry.beamline.optical_elements.absorbers.slit import WOSlit1D
from syned.beamline.element_coordinates import ElementCoordinates
from syned.beamline.beamline_element import BeamlineElement
from syned.beamline.shape import Rectangle

from srxraylib.plot.gol import plot


def initialize_default_propagator_1D():
    propagator = PropagationManager.Instance()

    propagator.add_propagator(Fraunhofer1D())
    propagator.add_propagator(Fresnel1D())
    propagator.add_propagator(FresnelConvolution1D())
    propagator.add_propagator(Integral1D())
    propagator.add_propagator(FresnelZoom1D())
    propagator.add_propagator(FresnelZoomScaling1D())


def propagate_wavefront(wavefront,distance,handler_name=Fresnel1D.HANDLER_NAME,zoom=0.005):

    wavefront_inside = wavefront.duplicate()

    slit = WOSlit1D(name="PIRRONE2",boundary_shape=Rectangle(-0.00005,0.00005,-0.00005,0.00005,))
    coordinates = ElementCoordinates(p = 0.0, q=0.0)
    propagation_elements = PropagationElements()
    propagation_elements.add_beamline_element(BeamlineElement(optical_element=slit,
                                                              coordinates=coordinates))
    parameters = PropagationParameters(wavefront=wavefront_inside,
                                       propagation_elements=propagation_elements)
    parameters.set_additional_parameters("shift_half_pixel", 1)
    parameters.set_additional_parameters("magnification_x", zoom)
    output_wf_1 = propagator.do_propagation(propagation_parameters=parameters,
                                            handler_name=handler_name)



    screen = WOScreen(name="PIRRONE")
    coordinates = ElementCoordinates(p = 0.0, q=distance)
    propagation_elements = PropagationElements()
    propagation_elements.add_beamline_element(BeamlineElement(optical_element=screen,
                                                              coordinates=coordinates))

    parameters = PropagationParameters(wavefront=output_wf_1,
                                       propagation_elements=propagation_elements)
    parameters.set_additional_parameters("shift_half_pixel", 1)
    parameters.set_additional_parameters("magnification_x", zoom)
    output_wf_2 = propagator.do_propagation(propagation_parameters=parameters,
                                            handler_name=handler_name)

    return output_wf_2


def compare(in_object_1,in_object_2,legend=["Spherical","Plane"]):


    xR = in_object_1.get_abscissas()
    yR = in_object_1.get_intensity()
    pR = in_object_1.get_phase(unwrap=1)


    xInf = in_object_2.get_abscissas()
    yInf = in_object_2.get_intensity()
    pInf = in_object_2.get_phase(unwrap=1)

    plot(xR,yR,xInf,yInf,legend=legend,title="Intensity")
    plot(xR,pR-pR.min(),xInf,pInf-pInf.min(),legend=legend,title="Phase")
    print("Total intensity: ",yR.sum())

#
#
#
#
#
#
def propagate_wavefront_scaling_new(wavefront,distance,handler_name=Fresnel1D.HANDLER_NAME,zoom=0.005,
                                using_radius=None):

    wavefront_inside = wavefront.duplicate()

    #
    # if R of curvature to be remove is not defined, guess it
    #
    if using_radius is None:
        plot_scan = True
        if plot_scan:
            radii,fig_of_mer = wavefront_inside.scan_wavefront_curvature(rmin=-1000,rmax=1000,rpoints=100)
            plot(radii,fig_of_mer)
        using_radius = wavefront_inside.guess_wavefront_curvature(rmin=-1000,rmax=1000,rpoints=100)

    #
    # propagate flattened wavefront to a new distance:  distance/magnification
    #

    slit = WOSlit1D(name="PIRRONE2",boundary_shape=Rectangle(-0.00005,0.00005,-0.00005,0.00005,))
    coordinates = ElementCoordinates(p = 0.0, q=0.0)
    propagation_elements = PropagationElements()
    propagation_elements.add_beamline_element(BeamlineElement(optical_element=slit,
                                                              coordinates=coordinates))
    parameters = PropagationParameters(wavefront=wavefront_inside,
                                       propagation_elements=propagation_elements)
    parameters.set_additional_parameters("shift_half_pixel", 1)
    parameters.set_additional_parameters("magnification_x", zoom)
    output_wf_1 = propagator.do_propagation(propagation_parameters=parameters,
                                            handler_name=handler_name)



    screen = WOScreen(name="PIRRONE")
    coordinates = ElementCoordinates(p = 0.0, q=distance)
    propagation_elements = PropagationElements()
    propagation_elements.add_beamline_element(BeamlineElement(optical_element=screen,
                                                              coordinates=coordinates))

    parameters = PropagationParameters(wavefront=output_wf_1,
                                       propagation_elements=propagation_elements)
    parameters.set_additional_parameters("shift_half_pixel", 1)
    parameters.set_additional_parameters("magnification_x", zoom)
    parameters.set_additional_parameters("radius", using_radius)
    output_wf_2 = propagator.do_propagation(propagation_parameters=parameters,
                                            handler_name=handler_name)


    return output_wf_2



if __name__ == "__main__":

    try:
        initialize_default_propagator_1D()
    except:
        print("Problems initializing 1D propagators")
        # pass

    propagator = PropagationManager.Instance()


    #
    # create source
    #
    wavefront = GenericWavefront1D.initialize_wavefront_from_range(x_min=-0.5e-3,
                                                                   x_max=0.5e-3,
                                                                   number_of_points=2048,
                                                                   wavelength=1.5e-10)
    radius = 50.0 # 50.
    wavefront.set_spherical_wave(radius=radius)


    distance = 5.0
    zoom = 1.0

    output_wf_2 = propagate_wavefront(wavefront,distance=distance,handler_name=FresnelZoom1D.HANDLER_NAME,zoom=zoom)
    output_wf_3 = propagate_wavefront_scaling_new(wavefront,distance=distance,handler_name=FresnelZoomScaling1D.HANDLER_NAME,
                                              zoom=zoom,using_radius=None)

    # plot(output_wf_2.get_abscissas()*1e6, output_wf_2.get_intensity(), title="WF2 - "+FresnelZoom1D.HANDLER_NAME)
    # plot(output_wf_3.get_abscissas()*1e6, output_wf_2.get_intensity(), title="SCALING!! WF2 - "+FresnelZoom1D.HANDLER_NAME)

    compare(output_wf_2,output_wf_3,legend=["zoom","zoom+scaling theorem"])



import unittest
import numpy
import os

from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D

from wofry.beamline.optical_elements.ideal_elements.lens import WOIdealLens, WOIdealLens1D

from wofry.propagator.polarization import Polarization

do_plot = False

#
# 1D tests
#
class GenericWavefront1DTest(unittest.TestCase):
    def test_initializers(self,do_plot=do_plot):

        print("#                                                             ")
        print("# Tests for initializars (1D)                                 ")
        print("#                                                             ")

        x = numpy.linspace(-100,100,50)
        y = numpy.abs(x)**1.5 +  1j*numpy.abs(x)**1.8



        wf0 = GenericWavefront1D.initialize_wavefront_from_steps(x[0],numpy.abs(x[1]-x[0]),y.size)
        wf0.set_complex_amplitude(y)

        wf1 = GenericWavefront1D.initialize_wavefront_from_range(x[0],x[-1],y.size)
        wf1.set_complex_amplitude(y)

        wf2 = GenericWavefront1D.initialize_wavefront_from_arrays(x,y)

        print("wavefront sizes: ",wf1.size(),wf1.size(),wf2.size())

        if do_plot:
            from srxraylib.plot.gol import plot
            plot(wf0.get_abscissas(),wf0.get_intensity(),
                       title="initialize_wavefront_from_steps",show=0)
            plot(wf1.get_abscissas(),wf1.get_intensity(),
                       title="initialize_wavefront_from_range",show=0)
            plot(wf2.get_abscissas(),wf2.get_intensity(),
                       title="initialize_wavefront_from_arrays",show=1)

        numpy.testing.assert_almost_equal(wf0.get_intensity(),numpy.abs(y)**2,5)
        numpy.testing.assert_almost_equal(wf1.get_intensity(),numpy.abs(y)**2,5)

        numpy.testing.assert_almost_equal(x,wf1.get_abscissas(),11)
        numpy.testing.assert_almost_equal(x,wf2.get_abscissas(),11)

    def test_plane_wave(self,do_plot=do_plot):
        #
        # plane wave
        #
        print("#                                                             ")
        print("# Tests for a 1D plane wave                                   ")
        print("#                                                             ")

        wavelength        = 1.24e-10

        wavefront_length_x = 400e-6

        npixels_x =  1024

        wavefront_x = numpy.linspace(-0.5*wavefront_length_x,0.5*wavefront_length_x,npixels_x)


        wavefront = GenericWavefront1D.initialize_wavefront_from_steps(
                        x_start=wavefront_x[0],x_step=numpy.abs(wavefront_x[1]-wavefront_x[0]),
                        number_of_points=npixels_x,wavelength=wavelength)

        numpy.testing.assert_almost_equal(wavefront_x,wavefront.get_abscissas(),9)

        # possible modifications

        wavefront.set_plane_wave_from_amplitude_and_phase(5.0,numpy.pi/2)
        numpy.testing.assert_almost_equal(wavefront.get_intensity(),25,5)

        wavefront.set_plane_wave_from_complex_amplitude(2.0+3j)
        numpy.testing.assert_almost_equal(wavefront.get_intensity(),13,5)

        phase_before = wavefront.get_phase()
        wavefront.add_phase_shift(numpy.pi/2)
        phase_after = wavefront.get_phase()
        numpy.testing.assert_almost_equal(phase_before+numpy.pi/2,phase_after,5)

        intensity_before = wavefront.get_intensity()
        wavefront.rescale_amplitude(10.0)
        intensity_after = wavefront.get_intensity()
        numpy.testing.assert_almost_equal(intensity_before*100,intensity_after,5)

        # interpolation

        wavefront.set_plane_wave_from_complex_amplitude(2.0+3j)
        test_value1 = wavefront.get_interpolated_complex_amplitude(0.01)
        self.assertAlmostEqual( (2.0+3j).real, test_value1.real, 5)
        self.assertAlmostEqual( (2.0+3j).imag, test_value1.imag, 5)


        if do_plot:
            from srxraylib.plot.gol import plot
            plot(wavefront.get_abscissas(),wavefront.get_intensity(),title="Intensity (plane wave)",show=0)
            plot(wavefront.get_abscissas(),wavefront.get_phase(),title="Phase (plane wave)",show=1)


    def test_spherical_wave(self,do_plot=do_plot):
        #
        # plane wave
        #
        print("#                                                             ")
        print("# Tests for a 1D spherical wave                               ")
        print("#                                                             ")

        wavelength        = 1.24e-10

        wavefront_length_x = 400e-6

        npixels_x =  1024

        wavefront_x = numpy.linspace(-0.5*wavefront_length_x,0.5*wavefront_length_x,npixels_x)



        wf1 = GenericWavefront1D.initialize_wavefront_from_steps(
                        x_start=wavefront_x[0],x_step=numpy.abs(wavefront_x[1]-wavefront_x[0]),
                        number_of_points=npixels_x,wavelength=wavelength)

        wf2 = GenericWavefront1D.initialize_wavefront_from_steps(
                        x_start=wavefront_x[0],x_step=numpy.abs(wavefront_x[1]-wavefront_x[0]),
                        number_of_points=npixels_x,wavelength=wavelength)

        # an spherical wavefront is obtained 1) by creation, 2) focusing a planewave

        wf1.set_spherical_wave(radius=-5.0, complex_amplitude=3+0j)
        wf1.clip(-50e-6,10e-6)

        wf2.set_plane_wave_from_complex_amplitude(3+0j)
        ideal_lens = WOIdealLens1D("test", 5.0)
        ideal_lens.applyOpticalElement(wf2)
        wf2.clip(-50e-6,10e-6)



        if do_plot:
            from srxraylib.plot.gol import plot
            plot(wf1.get_abscissas(),wf1.get_phase(),title="Phase of spherical wavefront",show=0)
            plot(wf2.get_abscissas(),wf2.get_phase(),title="Phase of focused plane wavefront",show=0)
            plot(wf1.get_abscissas(),wf1.get_phase(from_minimum_intensity=0.1),title="Phase of spherical wavefront (for intensity > 0.1)",show=0)
            plot(wf2.get_abscissas(),wf2.get_phase(from_minimum_intensity=0.1),title="Phase of focused plane wavefront (for intensity > 0.1)",show=1)


        numpy.testing.assert_almost_equal(wf1.get_phase(),wf2.get_phase(),5)

    def test_gaussianhermite_mode(self,do_plot=do_plot):
        #
        # plane wave
        #
        print("#                                                             ")
        print("# Tests for a 1D Gaussian Hermite mode                        ")
        print("#                                                             ")

        wavelength        = 1.24e-10


        # 2D
        sigma_x = 100e-6
        mode_x = 0
        npixels_x = 100


        wavefront_length_x = 10*sigma_x

        x = numpy.linspace(-0.5*wavefront_length_x,0.5*wavefront_length_x,npixels_x)


        wf1 = GenericWavefront1D.initialize_wavefront_from_steps(
                        x_start=x[0],x_step=numpy.abs(x[1]-x[0]),
                        number_of_points=npixels_x,wavelength=wavelength)


        wf1.set_gaussian_hermite_mode(sigma_x, mode_x, amplitude=1.0)


        numpy.testing.assert_almost_equal(wf1.get_amplitude()[30],23.9419082194,5)

        if do_plot:
            from srxraylib.plot.gol import plot
            plot(wf1.get_abscissas(),wf1.get_amplitude(),title="Amplitude of gaussianhermite",show=0)


    def test_interpolator(self,do_plot=do_plot):
        #
        # interpolator
        #
        print("#                                                             ")
        print("# Tests for 1D interpolator                                   ")
        print("#                                                             ")

        x = numpy.linspace(-10,10,100)

        sigma = 3.0
        Z = numpy.exp(-1.0*x**2/2/sigma**2)

        print("shape of Z",Z.shape)

        wf = GenericWavefront1D.initialize_wavefront_from_steps(x[0],numpy.abs(x[1]-x[0]),number_of_points=100)
        print("wf shape: ",wf.size())
        wf.set_complex_amplitude( Z )

        x1 = 3.2
        z1 = numpy.exp(x1**2/-2/sigma**2)
        print("complex ampl at (%g): %g+%gi (exact=%g)"%(x1,
                                                        wf.get_interpolated_complex_amplitude(x1).real,
                                                        wf.get_interpolated_complex_amplitude(x1).imag,
                                                        z1))
        self.assertAlmostEqual(wf.get_interpolated_complex_amplitude(x1).real,z1,4)

        print("intensity  at (%g):   %g (exact=%g)"%(x1,wf.get_interpolated_intensity(x1),z1**2))
        self.assertAlmostEqual(wf.get_interpolated_intensity(x1),z1**2,4)


        if do_plot:
            from srxraylib.plot.gol import plot
            plot(wf.get_abscissas(),wf.get_intensity(),title="Original",show=1)
            xx = wf.get_abscissas()
            yy = wf.get_interpolated_intensities(wf.get_abscissas()-1e-5)
            plot(xx,yy,title="interpolated on same grid",show=1)


    def test_save_load_h5_file(self):
        wfr = GenericWavefront1D.initialize_wavefront_from_range(-2.0,2.0,number_of_points=100)
        wfr.set_gaussian(.2,amplitude=5+8j)
        print("Saving 1D wavefront to file: tmp.h5")
        wfr.save_h5_file("tmp.h5","wfr1",intensity=True,phase=True)
        print("Reading 1D wavefront from file: tmp.h5")
        wfr2  = GenericWavefront1D.load_h5_file("tmp.h5","wfr1")
        print("Cleaning file tmp.h5")
        os.remove("tmp.h5")
        assert(wfr2.is_identical(wfr))

    def test_guess_wavefront_curvature(self):
        print("#                                                             ")
        print("# Tests guessing wavefront curvature                          ")
        print("#                                                             ")
        #
        # create source
        #
        wavefront = GenericWavefront1D.initialize_wavefront_from_range(x_min=-0.5e-3,
                                                                       x_max=0.5e-3,
                                                                       number_of_points=2048,
                                                                       wavelength=1.5e-10)
        radius = -50.0 # 50.
        wavefront.set_spherical_wave(radius=radius)
        if do_plot:
            from srxraylib.plot.gol import plot
            radii,fig_of_mer = wavefront.scan_wavefront_curvature(rmin=-1000,rmax=1000,rpoints=100)
            plot(radii,fig_of_mer)
        guess_radius = wavefront.guess_wavefront_curvature(rmin=-1000,rmax=1000,rpoints=100)
        assert(numpy.abs(radius - guess_radius) < 1e-3)

    def test_polarization(self):
        print("#                                                             ")
        print("# Tests polarization (1D)                                     ")
        print("#                                                             ")
        wavefront = GenericWavefront1D.initialize_wavefront_from_range(x_min=-0.5e-3,
                                                                       x_max=0.5e-3,
                                                                       number_of_points=2048,
                                                                       wavelength=1.5e-10,
                                                                       polarization=Polarization.TOTAL)

        ca = numpy.zeros(wavefront.size())
        wavefront.set_complex_amplitude(ca+(1+0j),ca+(0+1j))

        numpy.testing.assert_almost_equal(wavefront.get_interpolated_phase(0.1e-3,polarization=Polarization.SIGMA),0.0 )
        numpy.testing.assert_almost_equal(wavefront.get_interpolated_phase(-0.1e-3,polarization=Polarization.PI),numpy.pi/2 )

        numpy.testing.assert_almost_equal(wavefront.get_interpolated_intensities(-0.111e-3,polarization=Polarization.TOTAL),2.0 )
        numpy.testing.assert_almost_equal(wavefront.get_interpolated_intensities(-0.111e-3,polarization=Polarization.SIGMA),1.0 )
        numpy.testing.assert_almost_equal(wavefront.get_interpolated_intensities(-0.111e-3,polarization=Polarization.PI),1.0 )

        numpy.testing.assert_almost_equal(wavefront.get_intensity(polarization=Polarization.SIGMA),(ca+1)**2 )
        numpy.testing.assert_almost_equal(wavefront.get_intensity(polarization=Polarization.PI),   (ca+1)**2 )
        numpy.testing.assert_almost_equal(wavefront.get_intensity(polarization=Polarization.TOTAL),2*(ca+1)**2 )
#
# 2D tests
#

class GenericWavefront2DTest(unittest.TestCase):

    def test_initializers(self,do_plot=do_plot):

        print("#                                                             ")
        print("# Tests for initializars (2D)                                 ")
        print("#                                                             ")

        x = numpy.linspace(-100,100,50)
        y = numpy.linspace(-50,50,200)
        XY = numpy.meshgrid(x,y)
        X = XY[0].T
        Y = XY[1].T
        sigma = 10
        Z = numpy.exp(- (X**2 + Y**2)/2/sigma**2) * 1j
        print("Shapes x,y,z: ",x.shape,y.shape,Z.shape)

        wf0 = GenericWavefront2D.initialize_wavefront_from_steps(x[0],numpy.abs(x[1]-x[0]),y[0],numpy.abs(y[1]-y[0]),number_of_points=Z.shape)
        wf0.set_complex_amplitude(Z)

        wf1 = GenericWavefront2D.initialize_wavefront_from_range(x[0],x[-1],y[0],y[-1],number_of_points=Z.shape)
        wf1.set_complex_amplitude(Z)

        wf2 = GenericWavefront2D.initialize_wavefront_from_arrays(x,y,Z)

        if do_plot:
            from srxraylib.plot.gol import plot_image
            plot_image(wf0.get_intensity(),wf0.get_coordinate_x(),wf0.get_coordinate_y(),
                       title="initialize_wavefront_from_steps",show=0)
            plot_image(wf1.get_intensity(),wf1.get_coordinate_x(),wf1.get_coordinate_y(),
                       title="initialize_wavefront_from_range",show=0)
            plot_image(wf2.get_intensity(),wf2.get_coordinate_x(),wf2.get_coordinate_y(),
                       title="initialize_wavefront_from_arrays",show=1)


        numpy.testing.assert_almost_equal(numpy.abs(Z)**2,wf0.get_intensity(),11)
        numpy.testing.assert_almost_equal(numpy.abs(Z)**2,wf1.get_intensity(),11)
        numpy.testing.assert_almost_equal(numpy.abs(Z)**2,wf2.get_intensity(),11)

        numpy.testing.assert_almost_equal(x,wf0.get_coordinate_x(),11)
        numpy.testing.assert_almost_equal(x,wf1.get_coordinate_x(),11)
        numpy.testing.assert_almost_equal(x,wf2.get_coordinate_x(),11)

        numpy.testing.assert_almost_equal(y,wf0.get_coordinate_y(),11)
        numpy.testing.assert_almost_equal(y,wf1.get_coordinate_y(),11)
        numpy.testing.assert_almost_equal(y,wf2.get_coordinate_y(),11)



    def test_plane_wave(self,do_plot=do_plot):
        #
        # plane wave
        #
        print("#                                                             ")
        print("# Tests for a 2D plane wave                                      ")
        print("#                                                             ")

        wavelength        = 1.24e-10

        wavefront_length_x = 400e-6
        wavefront_length_y = wavefront_length_x

        npixels_x =  1024
        npixels_y =  npixels_x

        x = numpy.linspace(-0.5*wavefront_length_x,0.5*wavefront_length_x,npixels_x)
        y = numpy.linspace(-0.5*wavefront_length_y,0.5*wavefront_length_y,npixels_y)

        wavefront = GenericWavefront2D.initialize_wavefront_from_steps(
                        x_start=x[0],x_step=numpy.abs(x[1]-x[0]),
                        y_start=y[0],y_step=numpy.abs(y[1]-y[0]),
                        number_of_points=(npixels_x,npixels_y),wavelength=wavelength)

        # possible modifications

        wavefront.set_plane_wave_from_amplitude_and_phase(5.0,numpy.pi/2)
        numpy.testing.assert_almost_equal(wavefront.get_intensity(),25,5)

        wavefront.set_plane_wave_from_complex_amplitude(2.0+3j)
        numpy.testing.assert_almost_equal(wavefront.get_intensity(),13,5)

        phase_before = wavefront.get_phase()
        wavefront.add_phase_shift(numpy.pi/2)
        phase_after = wavefront.get_phase()
        numpy.testing.assert_almost_equal(phase_before+numpy.pi/2,phase_after,5)

        intensity_before = wavefront.get_intensity()
        wavefront.rescale_amplitude(10.0)
        intensity_after = wavefront.get_intensity()
        numpy.testing.assert_almost_equal(intensity_before*100,intensity_after,5)

        # interpolation

        wavefront.set_plane_wave_from_complex_amplitude(2.0+3j)
        test_value1 = wavefront.get_interpolated_complex_amplitude(0.01,1.3)
        self.assertAlmostEqual( (2.0+3j).real, test_value1.real, 5)
        self.assertAlmostEqual( (2.0+3j).imag, test_value1.imag, 5)


        if do_plot:
            from srxraylib.plot.gol import plot_image
            plot_image(wavefront.get_intensity(),wavefront.get_coordinate_x(),wavefront.get_coordinate_y(),
                       title="Intensity (plane wave)",show=0)
            plot_image(wavefront.get_phase(),wavefront.get_coordinate_x(),wavefront.get_coordinate_y(),
                       title="Phase (plane wave)",show=1)



    def test_spherical_wave(self,do_plot=do_plot):
        #
        # plane wave
        #
        print("#                                                             ")
        print("# Tests for a 2D spherical wave                               ")
        print("#                                                             ")

        wavelength        = 1.24e-10

        wavefront_length_x = 400e-6
        wavefront_length_y = wavefront_length_x

        npixels_x =  1024
        npixels_y =  npixels_x

        x = numpy.linspace(-0.5*wavefront_length_x,0.5*wavefront_length_x,npixels_x)
        y = numpy.linspace(-0.5*wavefront_length_y,0.5*wavefront_length_y,npixels_y)



        wf1 = GenericWavefront2D.initialize_wavefront_from_steps(
                        x_start=x[0],x_step=numpy.abs(x[1]-x[0]),
                        y_start=y[0],y_step=numpy.abs(y[1]-y[0]),
                        number_of_points=(npixels_x,npixels_y),wavelength=wavelength)

        numpy.testing.assert_almost_equal(x,wf1.get_coordinate_x(),9)
        numpy.testing.assert_almost_equal(y,wf1.get_coordinate_y(),9)

        wf2 = GenericWavefront2D.initialize_wavefront_from_steps(
                        x_start=x[0],x_step=numpy.abs(x[1]-x[0]),
                        y_start=y[0],y_step=numpy.abs(y[1]-y[0]),
                        number_of_points=(npixels_x,npixels_y),wavelength=wavelength)


        numpy.testing.assert_almost_equal(x,wf2.get_coordinate_x(),9)
        numpy.testing.assert_almost_equal(y,wf2.get_coordinate_y(),9)
        # an spherical wavefront is obtained 1) by creation, 2) focusing a planewave

        wf1.set_spherical_wave(-5.0, 3+0j)
        wf1.clip_square(-50e-6,10e-6,-20e-6,40e-6)

        wf2.set_plane_wave_from_complex_amplitude(3+0j)

        ideal_lens = WOIdealLens("test", 5.0, 5.0)
        ideal_lens.applyOpticalElement(wf2)

        wf2.clip_square(-50e-6,10e-6,-20e-6,40e-6)



        if do_plot:
            from srxraylib.plot.gol import plot_image
            plot_image(wf1.get_phase(),wf2.get_coordinate_x(),wf2.get_coordinate_y(),
                       title="Phase of spherical wavefront",show=0)
            plot_image(wf2.get_phase(),wf2.get_coordinate_x(),wf2.get_coordinate_y(),
                       title="Phase of focused plane wavefront",show=0)
            plot_image(wf1.get_phase(from_minimum_intensity=0.1),wf2.get_coordinate_x(),wf2.get_coordinate_y(),
                       title="Phase of spherical wavefront (for intensity > 0.1)",show=0)
            plot_image(wf2.get_phase(from_minimum_intensity=0.1),wf2.get_coordinate_x(),wf2.get_coordinate_y(),
                       title="Phase of focused plane wavefront (for intensity > 0.1)",show=1)


        numpy.testing.assert_almost_equal(wf1.get_phase(),wf2.get_phase(),5)

    def test_gaussianhermite_mode(self,do_plot=do_plot):
        #
        # plane wave
        #
        print("#                                                             ")
        print("# Tests for a 2D Gaussian Hermite mode                        ")
        print("#                                                             ")

        wavelength        = 1.24e-10


        # 2D
        sigma_x = 100e-6
        mode_x = 0
        npixels_x = 100
        sigma_y = 50e-6
        mode_y = 3
        npixels_y = 100


        wavefront_length_x = 10*sigma_x
        wavefront_length_y = 10*sigma_y

        x = numpy.linspace(-0.5*wavefront_length_x,0.5*wavefront_length_x,npixels_x)
        y = numpy.linspace(-0.5*wavefront_length_y,0.5*wavefront_length_y,npixels_y)



        wf1 = GenericWavefront2D.initialize_wavefront_from_steps(
                        x_start=x[0],x_step=numpy.abs(x[1]-x[0]),
                        y_start=y[0],y_step=numpy.abs(y[1]-y[0]),
                        number_of_points=(npixels_x,npixels_y),wavelength=wavelength)


        wf1.set_gaussian_hermite_mode(sigma_x, sigma_y, mode_x, mode_y, amplitude=1.0)


        numpy.testing.assert_almost_equal(wf1.get_amplitude()[30,40],1383.76448118,3)

        if do_plot:
            from srxraylib.plot.gol import plot_image
            plot_image(wf1.get_amplitude(),wf1.get_coordinate_x(),wf1.get_coordinate_y(),
                       title="Amplitude of gaussianhermite mode",show=1)



    def test_interpolator(self,do_plot=do_plot):
        #
        # interpolator
        #
        print("#                                                             ")
        print("# Tests for 2D interpolator                                   ")
        print("#                                                             ")

        x = numpy.linspace(-10,10,100)
        y = numpy.linspace(-20,20,50)

        xy = numpy.meshgrid(x,y)
        X = xy[0].T
        Y = xy[1].T
        sigma = 3.0
        Z = 3*numpy.exp(- (X**2+Y**2)/2/sigma**2) +4j

        print("shape of Z",Z.shape)

        wf = GenericWavefront2D.initialize_wavefront_from_steps(x[0],x[1]-x[0],y[0],y[1]-y[0],number_of_points=(100,50))
        print("wf shape: ",wf.size())
        print("wf polarized: ",wf.is_polarized())
        wf.set_complex_amplitude( Z )

        x1 = 3.2
        y1 = -2.5
        z1 = 3*numpy.exp(- (x1**2+y1**2)/2/sigma**2) + 4j
        print("complex ampl at (%g,%g): %g+%gi (exact=%g+%gi)"%(x1,y1,
                                                        wf.get_interpolated_complex_amplitude(x1,y1).real,
                                                        wf.get_interpolated_complex_amplitude(x1,y1).imag,
                                                        z1.real,z1.imag))
        self.assertAlmostEqual(wf.get_interpolated_complex_amplitude(x1,y1).real,z1.real,4)
        self.assertAlmostEqual(wf.get_interpolated_complex_amplitude(x1,y1).imag,z1.imag,4)
        #
        print("intensity  at (%g,%g):   %g (exact=%g)"%(x1,y1,wf.get_interpolated_intensity(x1,y1),numpy.abs(z1)**2))
        self.assertAlmostEqual(wf.get_interpolated_intensity(x1,y1),numpy.abs(z1)**2,3)

        # interpolate on same grid

        interpolated_complex_amplitude_on_same_grid = wf.get_interpolated_complex_amplitudes(X,Y)
        print("Shape interpolated at same grid: ",interpolated_complex_amplitude_on_same_grid.shape)

        numpy.testing.assert_array_almost_equal(wf.get_complex_amplitude(),interpolated_complex_amplitude_on_same_grid,4)

        print("Total intensity original wavefront: %g, interpolated on the same grid: %g"%
              (wf.get_intensity().sum(), (numpy.abs(interpolated_complex_amplitude_on_same_grid)**2).sum()))



        if do_plot:
            from srxraylib.plot.gol import plot_image
            plot_image(wf.get_intensity(),wf.get_coordinate_x(),wf.get_coordinate_y(),title="Original",show=0)
            plot_image(wf.get_interpolated_intensity(X,Y),wf.get_coordinate_x(),wf.get_coordinate_y(),
                       title="interpolated on same grid",show=1)


        # rebin wavefront

        # wf.set_plane_wave_from_complex_amplitude(3+4j)
        wf_rebin = wf.rebin(2.0,5.0,0.5,0.8,keep_the_same_intensity=1,set_extrapolation_to_zero=1)
        print("Shape before rebinning: ",wf.size())
        print("Shape after rebinning: ",wf_rebin.size())

        print("Total intensity original wavefront: %g, rebinned: %g"%
              (wf.get_intensity().sum(), wf_rebin.get_intensity().sum() ))

        if do_plot:
            from srxraylib.plot.gol import plot_image
            plot_image(wf.get_intensity(),wf.get_coordinate_x(),wf.get_coordinate_y(),title="BEFORE REBINNING",show=0)
            plot_image(wf_rebin.get_intensity(),wf_rebin.get_coordinate_x(),wf_rebin.get_coordinate_y(),
                       title="REBINNED",show=1)



    def test_save_load_h5_file(self):

        wfr = GenericWavefront2D.initialize_wavefront_from_range(-0.004,0.004,-0.001,0.001,(500,200))
        wfr.set_gaussian(0.002/6,0.001/12)
        wfr.save_h5_file("tmp_wofry.h5",subgroupname="wfr", intensity=True,phase=True,overwrite=True)

        wfr.set_gaussian(0.002/6/2,0.001/12/2)
        print("Writing file: tmp_wofry.h5")
        wfr.save_h5_file("tmp_wofry.h5",subgroupname="wfr2", intensity=True,phase=False,overwrite=False)

        # test same amplitudes:
        print("Accessing file, path: ","tmp_wofry.h5","wfr2")
        wfr2 = GenericWavefront2D.load_h5_file("tmp_wofry.h5","wfr2")
        print("Cleaning file tmp_wofry.h5")
        os.remove("tmp_wofry.h5")
        assert(wfr2.is_identical(wfr))

    def test_polarization(self):
        print("#                                                             ")
        print("# Tests polarization (2D)                                     ")
        print("#                                                             ")
        wavefront = GenericWavefront2D.initialize_wavefront_from_range(x_min=-0.5e-3,
                                                                       x_max=0.5e-3,
                                                                       y_min=-0.5e-3,
                                                                       y_max=0.5e-3,
                                                                       number_of_points=(2048,1024),
                                                                       wavelength=1.5e-10,
                                                                       polarization=Polarization.TOTAL)

        ca = numpy.zeros(wavefront.size())
        wavefront.set_complex_amplitude(ca+(1+0j),ca+(0+1j))
        #
        numpy.testing.assert_almost_equal(wavefront.get_interpolated_phase(0.1e-3,0.1e-3,polarization=Polarization.SIGMA),0.0 )
        numpy.testing.assert_almost_equal(wavefront.get_interpolated_phase(-0.1e-3,0.1e-3,polarization=Polarization.PI),numpy.pi/2 )

        numpy.testing.assert_almost_equal(wavefront.get_interpolated_intensities(-0.111e-3,-0.111e-3,polarization=Polarization.TOTAL),2.0 )
        numpy.testing.assert_almost_equal(wavefront.get_interpolated_intensities(-0.111e-3,-0.111e-3,polarization=Polarization.SIGMA),1.0 )
        numpy.testing.assert_almost_equal(wavefront.get_interpolated_intensities(-0.111e-3,-0.111e-3,polarization=Polarization.PI),1.0 )

        numpy.testing.assert_almost_equal(wavefront.get_intensity(polarization=Polarization.SIGMA),(ca+1)**2 )
        numpy.testing.assert_almost_equal(wavefront.get_intensity(polarization=Polarization.PI),   (ca+1)**2 )
        numpy.testing.assert_almost_equal(wavefront.get_intensity(polarization=Polarization.TOTAL),2*(ca+1)**2 )















import numpy

from srxraylib.util.data_structures import ScaledArray
from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
from wofry.propagator.propagator import Propagator1D

class FresnelConvolution1D(Propagator1D):

    HANDLER_NAME = "FRESNEL_CONVOLUTION_1D"

    def get_handler_name(self):
        return self.HANDLER_NAME

    def do_specific_progation_after(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    def do_specific_progation_before(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    """
    1D Fresnel propagator using direct convolution
    :param wavefront:
    :param propagation_distance:
    :return:
    """
    def do_specific_progation(self, wavefront, propagation_distance, parameters):
        # instead of numpy.convolve, this can be used:
        # from scipy.signal import fftconvolve

        kernel = numpy.exp(1j*2*numpy.pi/wavefront.get_wavelength() * wavefront.get_abscissas()**2 / 2 / propagation_distance)
        kernel *= numpy.exp(1j*2*numpy.pi/wavefront.get_wavelength() * propagation_distance)
        kernel /=  1j * wavefront.get_wavelength() * propagation_distance
        tmp = numpy.convolve(wavefront.get_complex_amplitude(),kernel,mode='same')

        wavefront_out =  GenericWavefront1D(wavefront.get_wavelength(), ScaledArray.initialize_from_steps(tmp,
                                    wavefront.offset(), wavefront.delta()))

        # added srio@esrf.eu 2018-03-23 to conserve energy - TODO: review method!
        wavefront_out.rescale_amplitude( numpy.sqrt(wavefront.get_intensity().sum() /
                                                    wavefront_out.get_intensity().sum()))

        return wavefront_out


#   propagate_2D_fresnel               \
#   propagate_2D_fresnel_convolution   | Near field Fresnel propagators via convolution in Fourier space. Three methods
#   propagate_2D_fresnel_srw          /
#
#          three methods available: 'fft': fft -> multiply by kernel in freq -> ifft
#                                   'convolution': scipy.signal.fftconvolve(wave,kernel in space)
#                                   'srw': use the SRW package
#
#
#
# *********************************** IMPORTANT *******************************************
#                RECOMMENDATIONS:
#
#     >>> Prefer propagate_2D_fresnel <<<
#       Prefer EVEN number of bins.
#       Set shift_half_pixel=1 (now the default)
#    Under these circumstances, the results agree very well with SRW
#
#
#


import numpy
import scipy.constants as codata

angstroms_to_eV = codata.h*codata.c/codata.e*1e10

from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D
from wofry.propagator.propagator import Propagator2D

class FresnelZoomXY2D(Propagator2D):

    HANDLER_NAME = "FRESNEL_ZOOM_XY_2D"

    def get_handler_name(self):
        return self.HANDLER_NAME

    def do_specific_progation_after(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    def do_specific_progation_before(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    """
    2D Fresnel propagator using convolution via Fourier transform
    :param wavefront:
    :param propagation_distance: propagation distance
    :param shift_half_pixel: set to 1 to shift half pixel (recommended using an even number of pixels) Set as default.
    :return: a new 2D wavefront object with propagated wavefront
    """

    def do_specific_progation(self, wavefront1, propagation_distance, parameters):
        if not parameters.has_additional_parameter("shift_half_pixel"):
            shift_half_pixel = True
        else:
            shift_half_pixel = parameters.get_additional_parameter("shift_half_pixel")

        wavefront = wavefront1.duplicate()
        shift_half_pixel = parameters.get_additional_parameter("shift_half_pixel")

        wavelength = wavefront.get_wavelength()


        wavenumber = wavefront.get_wavenumber()

        m_x = parameters.get_additional_parameter("magnification_x")
        m_y = parameters.get_additional_parameter("magnification_y")

        shape = wavefront.size()
        delta = wavefront.delta()

        pixelsize = delta[0]  # p_x[1] - p_x[0]
        npixels = shape[0]
        freq_nyquist = 0.5 / pixelsize
        freq_n = numpy.linspace(-1.0, 1.0, npixels)
        freq_x = freq_n * freq_nyquist

        # frequency for axis 2
        pixelsize = delta[1]
        npixels = shape[1]
        freq_nyquist = 0.5 / pixelsize
        freq_n = numpy.linspace(-1.0, 1.0, npixels)
        freq_y = freq_n * freq_nyquist

        if shift_half_pixel:
            freq_x = freq_x - 0.5 * numpy.abs(freq_x[1] - freq_x[0])
            freq_y = freq_y - 0.5 * numpy.abs(freq_y[1] - freq_y[0])

        f_x, f_y = numpy.meshgrid(freq_x, freq_y, indexing='ij')
        fsq = numpy.fft.fftshift(f_x ** 2 / m_x + f_y ** 2 / m_y)

        x = wavefront.get_mesh_x()
        y = wavefront.get_mesh_y()

        x_rescaling = wavefront.get_mesh_x() * m_x
        y_rescaling = wavefront.get_mesh_y() * m_y

        r1sq = x ** 2 * (1 - m_x) + y ** 2 * (1 - m_y)
        r2sq = x_rescaling ** 2 * ((m_x - 1) / m_x) + y_rescaling ** 2 * ((m_y - 1) / m_y)

        Q1 = wavenumber / 2 / propagation_distance * r1sq
        Q2 = numpy.exp(-1.0j * numpy.pi * wavelength * propagation_distance * fsq)
        Q3 = numpy.exp(1.0j * wavenumber / 2 / propagation_distance * r2sq)

        wavefront.add_phase_shift(Q1)

        fft = numpy.fft.fft2(wavefront.get_complex_amplitude())

        ifft = numpy.fft.ifft2(fft * Q2) * Q3 / numpy.sqrt(m_x * m_y)

        # wf_propagated = Wavefront2D.initialize_wavefront_from_arrays(wavefront.get_coordinate_x() * m_x,
        #                                                              ####################NON SONO SICURO CHE SIA GIUSTO QUELLO CHE RETURN
        #                                                              wavefront.get_coordinate_y() * m_y,
        #                                                              ifft,
        #                                                              wavelength=wavelength)
        #===============
        wf_propagated = GenericWavefront2D.initialize_wavefront_from_arrays(x_array=wavefront.get_coordinate_x()*m_x,
                                                                            y_array=wavefront.get_coordinate_y()*m_y,
                                                                            z_array=ifft,
                                                                            wavelength=wavelength)

        return wf_propagated

class FresnelConvolution2D(Propagator2D):

    HANDLER_NAME = "FRESNEL_CONVOLUTION_2D"

    def get_handler_name(self):
        return self.HANDLER_NAME


    def do_specific_progation_after(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    def do_specific_progation_before(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    """
    2D Fresnel propagator using convolution via Fourier transform
    :param wavefront:
    :param propagation_distance: propagation distance
    :param shift_half_pixel: set to 1 to shift half pixel (recommended using an even number of pixels) Set as default.
    :return: a new 2D wavefront object with propagated wavefront
    """

    def do_specific_progation(self, wavefront, propagation_distance, parameters):

        is_generic_wavefront = isinstance(wavefront, GenericWavefront2D)

        if is_generic_wavefront:
            pass
        else:
            wavefront_original = wavefront
            wavefront = wavefront.toGenericWavefront()

        if not parameters.has_additional_parameter("shift_half_pixel"):
            shift_half_pixel = True
        else:
            shift_half_pixel = parameters.get_additional_parameter("shift_half_pixel")

        from scipy.signal import fftconvolve

        wavelength = wavefront.get_wavelength()

        X = wavefront.get_mesh_x()
        Y = wavefront.get_mesh_y()

        if shift_half_pixel:
            x = wavefront.get_coordinate_x()
            y = wavefront.get_coordinate_y()
            X += 0.5 * numpy.abs( x[0] - x[1] )
            Y += 0.5 * numpy.abs( y[0] - y[1] )

        kernel = numpy.exp(1j*2*numpy.pi/wavefront.get_wavelength() *
                           (X**2 + Y**2) / 2 / propagation_distance)
        kernel *= numpy.exp(1j*2*numpy.pi/wavefront.get_wavelength() * propagation_distance)
        kernel /=  1j * wavefront.get_wavelength() * propagation_distance

        wf_propagated = GenericWavefront2D.initialize_wavefront_from_arrays(x_array=wavefront.get_coordinate_x(),
                                                                            y_array=wavefront.get_coordinate_y(),
                                                                            z_array=fftconvolve(wavefront.get_complex_amplitude(),
                                                                                                kernel,
                                                                                                mode='same'),
                                                                            wavelength=wavelength)
        return wf_propagated





import wpg.srwlib as srw

from wofry.propagator.wavefront import GenericWavefront1D, GenericWavefront2D, WavefrontDimension
from wofry.propagator.decorators import WavefrontDecorator
from wofry.propagator.propagator import AbstractPropagator, PropagationParameters

class WavefrontSRW(GenericWavefront2D, WavefrontDecorator):

    def __init__(self, parameteres=None):

        self.__srwwf =  srw.SRWLWfr()

        # do somethiing with paramaters

    def toGenericWavefront(self):
        # from SRW to Generic
        return GenericWavefront2D()

    def fromGenericWavefront(self, wavefront):
        # modify or replace self.__srwwf
        pass



class PropagatorSRW(AbstractPropagator):

    HANDLER = "SRW"

    def get_dimension(self):
        WavefrontDimension.TWO

    def get_handler_name(self):
        return self.HANDLER

    def do_propagation(self, parameters=PropagationParameters()):

        # build SRW Beamline and its parameters from list of elements

        # do calculation
        return WavefrontSRW()


class WavefrontWISE(GenericWavefront1D, WavefrontDecorator):

    def toGenericWavefront(self):
        # from SRW to Generic
        return GenericWavefront1D()

    def fromGenericWavefront(self, wavefront):
        # modify or replace self.__srwwf
        pass


# source = UndulatorSRW()
# srwwf = source.getWavefront()

srwwf = WavefrontSRW()

# do propagation stuff

#switch to WISE


wisewf = WavefrontWISE()
wisewf.fromGenericWavefront(srwwf.toGenericWavefront().get_Wavefront1D_from_histogram(axis=1))





# THIS FILE IS GENERATED FROM wofry SETUP.PY
short_version = '1.0.13'
version = '1.0.13'
full_version = '1.0.13'
git_revision = '3cc53b6a1b4f4b6aab5aae947bc48a93c0bffb0c'
release = True

if not release:
    version = full_version
    short_version += ".dev"



import numpy, copy
import scipy.constants as codata

from srxraylib.util.data_structures import ScaledArray, ScaledMatrix

from wofry.propagator.wavefront import Wavefront, WavefrontDimension

from wofry.propagator.util.gaussian_schell_model import GaussianSchellModel1D

# needed for h5 i/o
import os
import sys
import time
from wofry.propagator.polarization import Polarization
try:
    import h5py
except:
    raise ImportError("h5py not available: input/output to files not working")
# --------------------------------------------------
# Wavefront 1D
# --------------------------------------------------

class GenericWavefront1D(Wavefront):

    def __init__(self, wavelength=1e-10, electric_field_array=None, electric_field_array_pi=None):
        self._wavelength = wavelength
        self._electric_field_array = electric_field_array
        self._electric_field_array_pi = electric_field_array_pi

    def get_dimension(self):
        return WavefrontDimension.ONE

    def is_polarized(self):
        if self._electric_field_array_pi is None:
            return False
        else:
            return True

    def duplicate(self):
        if self.is_polarized():
            return GenericWavefront1D(wavelength=self._wavelength,
                                      electric_field_array=ScaledArray(np_array=copy.copy(self._electric_field_array.np_array),
                                                                       scale=copy.copy(self._electric_field_array.scale)),
                                      electric_field_array_pi=ScaledArray(np_array=copy.copy(self._electric_field_array_pi.np_array),
                                                                       scale=copy.copy(self._electric_field_array_pi.scale)) )
        else:
            return GenericWavefront1D(wavelength=self._wavelength,
                                      electric_field_array=ScaledArray(np_array=copy.copy(self._electric_field_array.np_array),
                                                                       scale=copy.copy(self._electric_field_array.scale)))

    @classmethod
    def initialize_wavefront(cls, wavelength=1e-10, number_of_points=1000, polarization=Polarization.SIGMA):

        sA = ScaledArray.initialize(np_array=numpy.full(number_of_points, (1.0 + 0.0j), dtype=complex))

        if ((polarization == Polarization.PI) or (polarization == Polarization.TOTAL)):
            sA_pi = ScaledArray.initialize(np_array=numpy.full(number_of_points, (0.0 + 0.0j), dtype=complex))
        else:
            sA_pi = None

        return GenericWavefront1D(wavelength, sA, sA_pi)

    @classmethod
    def initialize_wavefront_from_steps(cls, x_start=-1.0, x_step=0.002, number_of_points=1000, wavelength=1e-10, polarization=Polarization.SIGMA):

        sA = ScaledArray.initialize_from_steps(np_array=numpy.full(number_of_points, (1.0 + 0.0j), dtype=complex),
                                                                         initial_scale_value=x_start,
                                                                         scale_step=x_step)
        if ((polarization == Polarization.PI) or (polarization == Polarization.TOTAL)):
            sA_pi = ScaledArray.initialize_from_steps(np_array=numpy.full(number_of_points, (0.0 + 0.0j), dtype=complex),
                                                                         initial_scale_value=x_start,
                                                                         scale_step=x_step)
        else:
            sA_pi = None

        return GenericWavefront1D(wavelength, sA, sA_pi)

    @classmethod
    def initialize_wavefront_from_range(cls, x_min=0.0, x_max=0.0, number_of_points=1000, wavelength=1e-10, polarization=Polarization.SIGMA ):

        sA = ScaledArray.initialize_from_range(np_array=numpy.full(number_of_points, (1.0 + 0.0j), dtype=complex),
                                                                         min_scale_value=x_min,
                                                                         max_scale_value=x_max)

        if ((polarization == Polarization.PI) or (polarization == Polarization.TOTAL)):
            sA_pi = ScaledArray.initialize_from_range(np_array=numpy.full(number_of_points, (0.0 + 0.0j), dtype=complex),
                                                                             min_scale_value=x_min,
                                                                             max_scale_value=x_max)
        else:
            sA_pi = None
        return GenericWavefront1D(wavelength, sA, sA_pi )

    @classmethod
    def initialize_wavefront_from_arrays(cls, x_array, y_array, y_array_pi=None, wavelength=1e-10):
        if x_array.size != y_array.size:
            raise Exception("Unmatched shapes for x and y")

        sA = ScaledArray.initialize_from_steps(np_array=y_array,
                                               initial_scale_value=x_array[0],
                                               scale_step=numpy.abs(x_array[1]-x_array[0]))

        if y_array_pi is not None:
            sA_pi = ScaledArray.initialize_from_steps(np_array=y_array_pi,
                                                   initial_scale_value=x_array[0],
                                                   scale_step=numpy.abs(x_array[1]-x_array[0]))
        else:
            sA_pi = None

        return GenericWavefront1D(wavelength, sA, sA_pi)


    # main parameters

    # grid

    def size(self):
        return self._electric_field_array.size()

    def delta(self):
        return self._electric_field_array.delta()

    def offset(self):
        return self._electric_field_array.offset()

    def get_abscissas(self):
        return self._electric_field_array.scale

    def get_mesh_x(self):
        return self.get_abscissas()

    # photon energy

    def get_wavelength(self):
        return self._wavelength

    def get_wavenumber(self):
        return 2*numpy.pi/self._wavelength

    def get_photon_energy(self):
        m2ev = codata.c * codata.h / codata.e      # lambda(m)  = m2eV / energy(eV)
        return  m2ev / self._wavelength


    # wavefront

    def get_complex_amplitude(self, polarization=Polarization.SIGMA):

        if polarization == Polarization.SIGMA:
            return self._electric_field_array.np_array
        elif polarization == Polarization.PI:
            if self.is_polarized():
                return self._electric_field_array_pi.np_array
            else:
                raise Exception("Wavefront is not polarized.")
        else:
            raise Exception("Only 0=SIGMA and 1=PI are valid polarization values.")

    def get_amplitude(self, polarization=Polarization.SIGMA):
        return numpy.absolute(self.get_complex_amplitude(polarization=polarization))

    def get_phase(self,from_minimum_intensity=0.0,unwrap=0, polarization=Polarization.SIGMA):
        phase = numpy.angle(self.get_complex_amplitude(polarization=polarization))
        if (from_minimum_intensity > 0.0):
            intensity = self.get_intensity()
            intensity /= intensity.max()
            bad_indices = numpy.where(intensity < from_minimum_intensity )
            phase[bad_indices] = 0.0
        if unwrap:
            phase = numpy.unwrap(phase)
        return phase

    def get_intensity(self, polarization=Polarization.SIGMA):
        if polarization == Polarization.TOTAL:
            if self.is_polarized():
                return self.get_amplitude(polarization=Polarization.SIGMA)**2 + \
                       self.get_amplitude(polarization=Polarization.PI)**2
            else:
                return self.get_amplitude(polarization=Polarization.SIGMA)**2
        else:
            return self.get_amplitude(polarization=polarization)**2

    # interpolated values

    def get_interpolated_complex_amplitude(self, abscissa_value, polarization=Polarization.SIGMA): # singular

        if polarization == Polarization.SIGMA:
            return self._electric_field_array.interpolate_value(abscissa_value)
        elif polarization == Polarization.PI:
            if self.is_polarized():
                return self._electric_field_array_pi.interpolate_value(abscissa_value)
            else:
                raise Exception("Wavefront is not polarized.")
        else:
            raise Exception("Only 0=SIGMA and 1=PI are valid polarization values.")


    def get_interpolated_complex_amplitudes(self, abscissa_values, polarization=Polarization.SIGMA): # plural

        if polarization == Polarization.SIGMA:
            return self._electric_field_array.interpolate_values(abscissa_values)
        elif polarization == Polarization.PI:
            if self.is_polarized():
                return self._electric_field_array_pi.interpolate_values(abscissa_values)
            else:
                raise Exception("Wavefront is not polarized.")
        else:
            raise Exception("Only 0=SIGMA and 1=PI are valid polarization values.")



    def get_interpolated_amplitude(self, abscissa_value, polarization=Polarization.SIGMA): # singular!
        return numpy.absolute(self.get_interpolated_complex_amplitude(abscissa_value,polarization=polarization))

    def get_interpolated_amplitudes(self, abscissa_values, polarization=Polarization.SIGMA): # plural!
        return numpy.absolute(self.get_interpolated_complex_amplitudes(abscissa_values,polarization=polarization))

    def get_interpolated_phase(self, abscissa_value, polarization=Polarization.SIGMA): # singular!
        complex_amplitude = self.get_interpolated_complex_amplitude(abscissa_value, polarization=polarization)
        return numpy.arctan2(numpy.imag(complex_amplitude), numpy.real(complex_amplitude))

    def get_interpolated_phases(self, abscissa_values, polarization=Polarization.SIGMA): # plural!
        complex_amplitudes = self.get_interpolated_complex_amplitudes(abscissa_values, polarization=polarization)
        return numpy.arctan2(numpy.imag(complex_amplitudes), numpy.real(complex_amplitudes))

    def get_interpolated_intensity(self, abscissa_value, polarization=Polarization.SIGMA):
        if polarization == Polarization.TOTAL:
            interpolated_complex_amplitude = self.get_interpolated_amplitude(abscissa_value,polarization=Polarization.SIGMA)
            if self.is_polarized():
                interpolated_complex_amplitude_pi = self.get_interpolated_amplitude(abscissa_value,polarization=Polarization.PI)
                return numpy.abs(interpolated_complex_amplitude)**2 + numpy.abs(interpolated_complex_amplitude_pi)**2
            else:
                return numpy.abs(interpolated_complex_amplitude)**2
        elif polarization == Polarization.SIGMA:
            interpolated_complex_amplitude = self.get_interpolated_amplitude(abscissa_value,polarization=Polarization.SIGMA)
            return numpy.abs(interpolated_complex_amplitude)**2
        elif polarization == Polarization.PI:
            interpolated_complex_amplitude_pi = self.get_interpolated_amplitude(abscissa_value,polarization=Polarization.PI)
            return numpy.abs(interpolated_complex_amplitude_pi)**2
        else:
            raise Exception("Wrong polarization value.")


    def get_interpolated_intensities(self, abscissa_values, polarization=Polarization.SIGMA):
        # return self.get_interpolated_amplitudes(abscissa_values,polarization=Polarization.SIGMA)**2

        if polarization == Polarization.TOTAL:
            interpolated_complex_amplitude = self.get_interpolated_complex_amplitude(abscissa_values,polarization=Polarization.SIGMA)
            if self.is_polarized():
                interpolated_complex_amplitude_pi = self.get_interpolated_complex_amplitude(abscissa_values,polarization=Polarization.PI)
                return numpy.abs(interpolated_complex_amplitude)**2 + numpy.abs(interpolated_complex_amplitude_pi)**2
            else:
                return numpy.abs(interpolated_complex_amplitude)**2
        elif polarization == Polarization.SIGMA:
            interpolated_complex_amplitude = self.get_interpolated_complex_amplitude(abscissa_values,polarization=Polarization.SIGMA)
            return numpy.abs(interpolated_complex_amplitude)**2
        elif polarization == Polarization.PI:
            interpolated_complex_amplitude_pi = self.get_interpolated_complex_amplitude(abscissa_values,polarization=Polarization.PI)
            return numpy.abs(interpolated_complex_amplitude_pi)**2
        else:
            raise Exception("Wrong polarization value.")


    # modifiers

    def set_wavelength(self, wavelength):
        self._wavelength = wavelength

    def set_wavenumber(self, wavenumber):
        self._wavelength = 2 * numpy.pi / wavenumber

    def set_photon_energy(self, photon_energy):
        m2ev = codata.c * codata.h / codata.e      # lambda(m)  = m2eV / energy(eV)
        self._wavelength = m2ev / photon_energy

    def set_complex_amplitude(self, complex_amplitude, complex_amplitude_pi=None):
        if complex_amplitude.size != self._electric_field_array.size():
            raise Exception("Complex amplitude array has different dimension")

        self._electric_field_array.np_array = complex_amplitude

        if complex_amplitude_pi is not None:
            if self.is_polarized():
                if complex_amplitude_pi.size != self._electric_field_array_pi.size():
                    raise Exception("Complex amplitude array has different dimension")

                self._electric_field_array_pi.np_array = complex_amplitude_pi
            else:
                raise Exception('Cannot set PI-polarized complex amplitude to a non-polarized wavefront.')

    def set_pi_complex_amplitude_to_zero(self):
        if self.is_polarized():
            self._electric_field_array_pi.np_array *= 0.0

    def set_plane_wave_from_complex_amplitude(self, complex_amplitude=(1.0 + 0.0j), inclination=0.0):
        self._electric_field_array.np_array = numpy.full(self._electric_field_array.size(), complex_amplitude, dtype=complex)
        if inclination != 0.0:
            self.add_phase_shifts( self.get_wavenumber() * self._electric_field_array.scale * numpy.tan(inclination) )
        # if polarized, set arbitrary PI component to zero
        self.set_pi_complex_amplitude_to_zero()

    def set_plane_wave_from_amplitude_and_phase(self, amplitude=1.0, phase=0.0, inclination=0.0):
        self.set_plane_wave_from_complex_amplitude(amplitude*numpy.cos(phase) + 1.0j*amplitude*numpy.sin(phase))
        if inclination != 0.0:
            self.add_phase_shifts( self.get_wavenumber() * self._electric_field_array.scale * numpy.tan(inclination) )
        # if polarized, set arbitrary PI component to zero
        self.set_pi_complex_amplitude_to_zero()

    def set_spherical_wave(self, radius=1.0, center=0.0, complex_amplitude=1.0):
        if radius == 0: raise Exception("Radius cannot be zero")

        self._electric_field_array.np_array = complex_amplitude * numpy.exp(-1.0j * self.get_wavenumber() *
                                            ( (self._electric_field_array.scale - center)** 2) / (-2 * radius))
        # if polarized, set arbitrary PI component to zero
        self.set_pi_complex_amplitude_to_zero()

    def set_gaussian_hermite_mode(self, sigma_x, mode_x, amplitude=1.0, shift=0.0):
        a1D = GaussianSchellModel1D(amplitude, sigma_x, 100.0*sigma_x)

        real_amplitude = a1D.phi(mode_x, self.get_abscissas() - shift)

        self.set_complex_amplitude(real_amplitude+0.0j)
        # if polarized, set arbitrary PI component to zero
        self.set_pi_complex_amplitude_to_zero()

    # note that amplitude is for "amplitude" not for intensity!
    def set_gaussian(self, sigma_x, amplitude=1.0, shift=0.0):
        self.set_gaussian_hermite_mode(sigma_x, 0, amplitude=amplitude, shift=shift)
        # if polarized, set arbitrary PI component to zero
        self.set_pi_complex_amplitude_to_zero()

    def add_phase_shift(self, phase_shift, polarization=Polarization.SIGMA):
        if polarization == Polarization.SIGMA:
            self._electric_field_array.np_array *= numpy.exp(1.0j * phase_shift)
        elif polarization == Polarization.PI:
            if self.is_polarized():
                self._electric_field_array_pi.np_array *= numpy.exp(1.0j * phase_shift)
            else:
                raise Exception("Wavefront is not polarized")
        else:
            raise Exception("Invalid polarization value (only 0=SIGMA or 1=PI are valid)")

    def add_phase_shifts(self, phase_shifts, polarization=Polarization.SIGMA):

        if polarization == Polarization.SIGMA:
            if phase_shifts.size != self._electric_field_array.size():
                raise Exception("Phase Shifts array has different dimension")
            self._electric_field_array.np_array =  numpy.multiply(self._electric_field_array.np_array, numpy.exp(1.0j * phase_shifts))
        elif polarization == Polarization.PI:
            if self.is_polarized():
                if phase_shifts.size != self._electric_field_array_pi.size():
                    raise Exception("Phase Shifts array has different dimension")
                self._electric_field_array_pi.np_array =  numpy.multiply(self._electric_field_array_pi.np_array, numpy.exp(1.0j * phase_shifts))
            else:
                raise Exception("Wavefront is not polarized")
        else:
            raise Exception("Invalid polarization value (only 0=SIGMA or 1=PI are valid)")


    def rescale_amplitude(self, factor, polarization=Polarization.SIGMA):

        if polarization == Polarization.SIGMA:
            self._electric_field_array.np_array *= factor
        elif polarization == Polarization.PI:
            if self.is_polarized():
                self._electric_field_array_pi.np_array *= factor
            else:
                raise Exception("Wavefront is not polarized")
        elif polarization == Polarization.TOTAL:
            self.rescale_amplitude(factor, polarization=Polarization.SIGMA)
            self.rescale_amplitude(factor, polarization=Polarization.PI)
        else:
            raise Exception("Invalid polarization value (only 0=SIGMA, 1=PI or 3=TOTAL are valid)")


    def rescale_amplitudes(self, factors, polarization=Polarization.SIGMA):


        if polarization == Polarization.SIGMA:
            if factors.size != self._electric_field_array.size(): raise Exception("Factors array has different dimension")
            self._electric_field_array.np_array =  numpy.multiply(self._electric_field_array.np_array, factors)
        elif polarization == Polarization.PI:
            if self.is_polarized():
                if factors.size != self._electric_field_array_pi.size(): raise Exception("Factors array has different dimension")
                self._electric_field_array_pi.np_array =  numpy.multiply(self._electric_field_array_pi.np_array, factors)
            else:
                raise Exception("Wavefront is not polarized")
        elif polarization == Polarization.TOTAL:
            self.rescale_amplitudes(factors, polarization=Polarization.SIGMA)
            self.rescale_amplitudes(factors, polarization=Polarization.PI)
        else:
            raise Exception("Invalid polarization value (only 0=SIGMA, 1=PI or 3=TOTAL are valid)")


    def clip(self, x_min, x_max, negative=False):
        window = numpy.ones(self._electric_field_array.size())

        if not negative:
            lower_window = numpy.where(self.get_abscissas() < x_min)
            upper_window = numpy.where(self.get_abscissas() > x_max)

            if len(lower_window) > 0: window[lower_window] = 0
            if len(upper_window) > 0: window[upper_window] = 0
        else:
            window_indices = numpy.where((self.get_abscissas() >= x_min) & (self.get_abscissas() <= x_max))

            if len(window_indices) > 0:
                window[window_indices] = 0.0

        if self.is_polarized():
            self.rescale_amplitudes(window,polarization=Polarization.TOTAL)
        else:
            self.rescale_amplitudes(window,polarization=Polarization.SIGMA)

    def is_identical(self,wfr,decimal=7):
        from numpy.testing import assert_array_almost_equal
        try:
            assert_array_almost_equal(self.get_complex_amplitude(),wfr.get_complex_amplitude(),decimal)
            assert(self.is_polarized() == wfr.is_polarized())
            if self.is_polarized():
                assert_array_almost_equal(self.get_complex_amplitude(polarization=Polarization.PI),
                                          wfr.get_complex_amplitude(polarization=Polarization.PI),decimal)

            assert_array_almost_equal(self.get_abscissas(),wfr.get_abscissas(),decimal)
            assert_array_almost_equal(self.get_photon_energy(),wfr.get_photon_energy(),decimal)
        except:
            return False

        return True


    #
    # auxiliary methods get main wavefront phase curvature (radius)
    #
    def _figure_of_merit(self,radius,weight_with_intensity=True):
        """
        Computes a "figure of merit" for finding the wavefront curvature.
        A low value of the figure of metit means that the entered radius (checked)
        corresponds to the radius of the wavefront.

        If wavefront is polarized, the pi component is ignored.

        :param radius:
        :param weight_with_intensity:
        :return: a positive scalar with the figure of merit
        """
        x = self.get_abscissas()
        new_phase = 1.0 * self.get_wavenumber() * (x**2) / (-2 * radius)

        wavefront2 = self.duplicate()
        wavefront2.add_phase_shifts(new_phase)

        if weight_with_intensity:
            out = numpy.abs(wavefront2.get_phase()*wavefront2.get_intensity()).sum()
        else:
            out = numpy.abs(wavefront2.get_phase()).sum()

        return out

    def scan_wavefront_curvature(self,rmin=-10000.0,rmax=10000.0,rpoints=100):

        radii = numpy.linspace(rmax,rmin,rpoints)
        fig_of_mer = numpy.zeros_like(radii)

        for i,radius in enumerate(radii):
            fig_of_mer[i] =self._figure_of_merit(radius)

        return radii,fig_of_mer


    def guess_wavefront_curvature(self,rmin=-10000.0,rmax=10000.0,rpoints=100):
        from scipy.optimize import minimize

        radii,fig_of_mer = self.scan_wavefront_curvature(rmin=rmin,rmax=rmax,rpoints=rpoints)

        res = minimize(self._figure_of_merit, radii[numpy.argmin(fig_of_mer)], args=self, method='powell',options={'xtol': 1e-8, 'disp': True})

        return res.x

    #
    # auxiliary function to dump h5 files
    #
    def _dump_arr_2_hdf5(self,_arr,_calculation, _filename, _subgroupname):
        """
        Auxiliary routine to save_h5_file
        :param _arr: (usually 2D) array to be saved on the hdf5 file inside the _subgroupname
        :param _calculation
        :param _filename: path to file for saving the wavefront
        :param _subgroupname: container mechanism by which HDF5 files are organised
        """
        sys.stdout.flush()
        f = h5py.File(_filename, 'a')
        try:
            f1 = f.create_group(_subgroupname)
        except:
            f1 = f[_subgroupname]
        fdata = f1.create_dataset(_calculation, data=_arr)
        f.close()

    def save_h5_file(self,filename,subgroupname="wfr",intensity=False,phase=False,overwrite=True,verbose=False):
        """
        Auxiliary function to write wavefront data into a hdf5 generic file.
        When using the append mode to write h5 files, overwriting does not work and makes the code crash. To avoid this
        issue, try/except is used. If by any chance a file should be overwritten, it is firstly deleted and re-written.
        :param self: input / output resulting Wavefront structure (instance of GenericWavefront2D);
        :param filename: path to file for saving the wavefront
        :param subgroupname: container mechanism by which HDF5 files are organised
        :param intensity: writes intensity for sigma and pi polarisation (default=False)
        :param amplitude:
        :param phase:
        :param overwrite: flag that should always be set to True to avoid infinity loop on the recursive part of the function.
        :param verbose: if True, print some file i/o messages
        """
        try:
            if not os.path.isfile(filename):  # if file doesn't exist, create it.
                sys.stdout.flush()
                f = h5py.File(filename, 'w')
                # point to the default data to be plotted
                f.attrs['default']          = 'entry'
                # give the HDF5 root some more attributes
                f.attrs['file_name']        = filename
                f.attrs['file_time']        = time.time()
                f.attrs['creator']          = 'oasys-wofry'
                f.attrs['HDF5_Version']     = h5py.version.hdf5_version
                f.attrs['h5py_version']     = h5py.version.version
                f.close()

            # always writes complex amplitude
            x_polarization = self.get_complex_amplitude()       # sigma
            self._dump_arr_2_hdf5(x_polarization, "wfr_complex_amplitude_s", filename, subgroupname)
            self._dump_arr_2_hdf5(x_polarization, "wfr_complex_amplitude_p", filename, subgroupname)

            if self.is_polarized():
                y_polarization = self.get_complex_amplitude(polarization=Polarization.PI)       # pi
                self._dump_arr_2_hdf5(y_polarization, "wfr_complex_amplitude_p", filename, subgroupname)



            if intensity:

                if self.is_polarized():
                    self._dump_arr_2_hdf5(self.get_intensity(polarization=Polarization.TOTAL),"intensity/wfr_intensity", filename, subgroupname)
                    self._dump_arr_2_hdf5(self.get_intensity(polarization=Polarization.SIGMA),"intensity/wfr_intensity_s", filename, subgroupname)
                    self._dump_arr_2_hdf5(self.get_intensity(polarization=Polarization.PI),"intensity/wfr_intensity_p", filename, subgroupname)
                else:
                    self._dump_arr_2_hdf5(self.get_intensity(),"intensity/wfr_intensity", filename, subgroupname)

            if phase:
                if self.is_polarized():
                    self._dump_arr_2_hdf5(self.get_phase(polarization=Polarization.SIGMA)-self.get_phase(polarization=Polarization.PI),
                                          "phase/wfr_phase", filename, subgroupname)
                    self._dump_arr_2_hdf5(self.get_phase(polarization=Polarization.SIGMA),"phase/wfr_phase_s", filename, subgroupname)
                    self._dump_arr_2_hdf5(self.get_phase(polarization=Polarization.PI),"phase/wfr_phase_p", filename, subgroupname)
                else:
                    self._dump_arr_2_hdf5(self.get_phase(polarization=Polarization.SIGMA),"phase/wfr_phase", filename, subgroupname)


            # add mesh and SRW information
            f = h5py.File(filename, 'a')
            f1 = f[subgroupname]

            # point to the default data to be plotted
            f1.attrs['NX_class'] = 'NXentry'
            f1.attrs['default'] = 'intensity'

            # TODO: add self interpreting decoder
            # f1["wfr_method"] = "WOFRY"
            f1["wfr_dimension"] = 1
            f1["wfr_photon_energy"] = self.get_photon_energy()
            x = self.get_abscissas()


            f1["wfr_mesh"] =  numpy.array([x[0],x[-1],x.size])


            # Add NX plot attribites for automatic plot with silx view
            myflags = [intensity,phase]
            mylabels = ['intensity','phase']
            for i,label in enumerate(mylabels):
# nxdata = nxentry.create_group('mr_scan')
# nxdata.attrs['NX_class'] = 'NXdata'
# nxdata.attrs['signal'] = 'I00'      # Y axis of default plot
# nxdata.attrs['axes'] = 'mr'         # X axis of default plot
# nxdata.attrs['mr_indices'] = [0,]   # use "mr" as the first dimension of I00
                if myflags[i]:
                    f2 = f1[mylabels[i]]
                    f2.attrs['NX_class'] = 'NXdata'
                    f2.attrs['signal'] = 'wfr_%s'%(mylabels[i])
                    f2.attrs['axes'] = b'axis_x'

                    f3 = f2["wfr_%s"%(mylabels[i])]
                    # f3.attrs['interpretation'] = 'image'


                    # axis data
                    ds = f2.create_dataset('axis_x', data=1e6*x)
                    ds.attrs['units'] = 'microns'
                    ds.attrs['long_name'] = 'Pixel Size (microns)'    # suggested Y axis plot label
            f.close()

        except:
            # TODO: check exit??
            if overwrite is not True:
                raise Exception("Bad input argument")
            os.remove(filename)
            if verbose: print("save_h5_file: file deleted %s"%filename)
            self.save_h5_file(filename,subgroupname, intensity=intensity, phase=phase, overwrite=False)

        if verbose: print("save_h5_file: written/updated %s data in file: %s"%(subgroupname,filename))

    @classmethod
    def load_h5_file(cls,filename,filepath):

        try:
            f = h5py.File(filename, 'r')
            mesh = f[filepath+"/wfr_mesh"].value

            complex_amplitude_s = f[filepath+"/wfr_complex_amplitude_s"].value
            wfr = cls.initialize_wavefront_from_arrays(
                                numpy.linspace(mesh[0],mesh[1],mesh[2]),
                                complex_amplitude_s)
            wfr.set_photon_energy(f[filepath+"/wfr_photon_energy"].value)
            f.close()
            return wfr
        except:
            raise Exception("Failed to load 2D wavefront to h5 file: "+filename)

if __name__ == "__main__":
    # w = GenericWavefront1D.initialize_wavefront_from_steps()
    # w2 = w.duplicate()


    pass




import numpy

from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
from wofry.propagator.propagator import Propagator1D, PropagationParameters, PropagationElements

class Fraunhofer1D(Propagator1D):

    HANDLER_NAME = "FRAUNHOFER_1D"

    def get_handler_name(self):
        return self.HANDLER_NAME

    def do_specific_progation_after(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    def do_specific_progation_before(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    """
    1D Fraunhofer propagator using convolution via Fourier transform
    :param wavefront:
    :param propagation_distance: propagation distance. If set to zero, the abscissas
                                 of the returned wavefront are in angle (rad)
    :return: a new 1D wavefront object with propagated wavefront
    """

    # TODO: check resulting amplitude normalization
    def do_specific_progation(self, wavefront, propagation_distance, parameters):

        shape = wavefront.size()
        delta = wavefront.delta()
        wavelength = wavefront.get_wavelength()
        wavenumber = wavefront.get_wavenumber()
        fft_scale = numpy.fft.fftfreq(shape, d=delta)
        fft_scale = numpy.fft.fftshift(fft_scale)
        x2 = fft_scale * propagation_distance * wavelength

        try:
            if parameters.get_additional_parameter("shift_half_pixel"):
                x2 = x2 - 0.5 * numpy.abs(x2[1] - x2[0])
        except:
            pass


        p1 = numpy.exp(1.0j * wavenumber * propagation_distance)
        p2 = numpy.exp(1.0j * wavenumber / 2 / propagation_distance * x2**2)
        p3 = 1.0j*wavelength*propagation_distance

        fft = numpy.fft.fft(wavefront.get_complex_amplitude())
        fft = fft * p1 * p2 / p3
        fft2 = numpy.fft.fftshift(fft)

        wavefront_out =  GenericWavefront1D.initialize_wavefront_from_arrays(x2, fft2, wavelength=wavefront.get_wavelength())


        # added srio@esrf.eu 2018-03-23 to conserve energy - TODO: review method!
        wavefront_out.rescale_amplitude( numpy.sqrt(wavefront.get_intensity().sum() /
                                                    wavefront_out.get_intensity().sum()))

        return wavefront_out


#   propagate_2D_integral: Simplification of the Kirchhoff-Fresnel integral. TODO: Very slow and give some problems

import numpy

from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D
from wofry.propagator.propagator import Propagator2D

# TODO: check resulting amplitude normalization (fft and srw likely agree, convolution gives too high amplitudes, so needs normalization)

class Integral2D(Propagator2D):

    HANDLER_NAME = "INTEGRAL_2D"

    def get_handler_name(self):
        return self.HANDLER_NAME

    def do_specific_progation_after(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    def do_specific_progation_before(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    """
    2D Fresnel-Kirchhoff propagator via simplified integral

    NOTE: this propagator is experimental and much less performant than the ones using Fourier Optics
          Therefore, it is not recommended to use.

    :param wavefront:
    :param propagation_distance: propagation distance
    :param shuffle_interval: it is known that this method replicates the central diffraction spot
                            The distace of the replica is proportional to 1/pixelsize
                            To avoid that, it is possible to change a bit (randomly) the coordinates
                            of the wavefront. shuffle_interval controls this shift: 0=No shift. A typical
                             value can be 1e5.
                             The result shows a diffraction pattern without replica but with much noise.
    :param calculate_grid_only: if set, it calculates only the horizontal and vertical profiles, but returns the
                             full image with the other pixels to zero. This is useful when calculating large arrays,
                             so it is set as the default.
    :return: a new 2D wavefront object with propagated wavefront
    """

    def do_specific_progation(self, wavefront, propagation_distance, parameters):
        if not parameters.has_additional_parameter("shuffle_interval"):
            shuffle_interval = 0
        else:
            shuffle_interval = parameters.get_additional_parameter("shuffle_interval")

        if not parameters.has_additional_parameter("calculate_grid_only"):
            calculate_grid_only = 1
        else:
            calculate_grid_only = parameters.get_additional_parameter("calculate_grid_only")

        #
        # Fresnel-Kirchhoff integral (neglecting inclination factor)
        #

        if calculate_grid_only == 0:
            #
            # calculation over the whole detector area
            #
            p_x = wavefront.get_coordinate_x()
            p_y = wavefront.get_coordinate_y()
            wavelength = wavefront.get_wavelength()
            amplitude = wavefront.get_complex_amplitude()

            det_x = p_x.copy()
            det_y = p_y.copy()

            p_X = wavefront.get_mesh_x()
            p_Y = wavefront.get_mesh_y()

            det_X = p_X
            det_Y = p_Y


            amplitude_propagated = numpy.zeros_like(amplitude,dtype='complex')

            wavenumber = 2 * numpy.pi / wavelength

            for i in range(det_x.size):
                for j in range(det_y.size):
                    if shuffle_interval == 0:
                        rd_x = 0.0
                        rd_y = 0.0
                    else:
                        rd_x = (numpy.random.rand(p_x.size,p_y.size)-0.5)*shuffle_interval
                        rd_y = (numpy.random.rand(p_x.size,p_y.size)-0.5)*shuffle_interval

                    r = numpy.sqrt( numpy.power(p_X + rd_x - det_X[i,j],2) +
                                    numpy.power(p_Y + rd_y - det_Y[i,j],2) +
                                    numpy.power(propagation_distance,2) )

                    amplitude_propagated[i,j] = (amplitude / r * numpy.exp(1.j * wavenumber *  r)).sum()

            output_wavefront = GenericWavefront2D.initialize_wavefront_from_arrays(det_x,det_y,amplitude_propagated)

        else:
            x = wavefront.get_coordinate_x()
            y = wavefront.get_coordinate_y()
            X = wavefront.get_mesh_x()
            Y = wavefront.get_mesh_y()
            wavenumber = 2 * numpy.pi / wavefront.get_wavelength()
            amplitude = wavefront.get_complex_amplitude()

            used_indices = wavefront.get_mask_grid(width_in_pixels=(1,1),number_of_lines=(1,1))
            indices_x = wavefront.get_mesh_indices_x()
            indices_y = wavefront.get_mesh_indices_y()

            indices_x_flatten = indices_x[numpy.where(used_indices == 1)].flatten()
            indices_y_flatten = indices_y[numpy.where(used_indices == 1)].flatten()
            X_flatten         =         X[numpy.where(used_indices == 1)].flatten()
            Y_flatten         =         Y[numpy.where(used_indices == 1)].flatten()
            complex_amplitude_propagated = amplitude*0

            print("propagate_2D_integral: Calculating %d points from a total of %d x %d = %d"%(
                X_flatten.size,amplitude.shape[0],amplitude.shape[1],amplitude.shape[0]*amplitude.shape[1]))

            for i in range(X_flatten.size):
                r = numpy.sqrt( numpy.power(wavefront.get_mesh_x() - X_flatten[i],2) +
                                numpy.power(wavefront.get_mesh_y() - Y_flatten[i],2) +
                                numpy.power(propagation_distance,2) )

                complex_amplitude_propagated[int(indices_x_flatten[i]),int(indices_y_flatten[i])] = (amplitude / r * numpy.exp(1.j * wavenumber *  r)).sum()

            output_wavefront = GenericWavefront2D.initialize_wavefront_from_arrays(x_array=x,
                                                                                   y_array=y,
                                                                                   z_array=complex_amplitude_propagated,
                                                                                   wavelength=wavefront.get_wavelength())

        # added srio@esrf.eu 2018-03-23 to conserve energy - TODO: review method!
        output_wavefront.rescale_amplitude( numpy.sqrt(wavefront.get_intensity().sum() /
                                                    output_wavefront.get_intensity().sum()))

        return output_wavefront



import numpy

from srxraylib.util.data_structures import ScaledArray
from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
from wofry.propagator.propagator import Propagator1D

class Integral1D(Propagator1D):

    HANDLER_NAME = "INTEGRAL_1D"

    def get_handler_name(self):
        return self.HANDLER_NAME

    def do_specific_progation_after(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    def do_specific_progation_before(self, wavefront, propagation_distance, parameters):
        return self.do_specific_progation(wavefront, propagation_distance, parameters)

    # 1D Fresnel-Kirchhoff propagator via simplified integral
    def do_specific_progation(self, wavefront, propagation_distance, parameters):

        try:
            mX = parameters.get_additional_parameter("magnification_x")
        except:
            mX = 1.0

        try:
            mN = parameters.get_additional_parameter("magnification_N")
        except:
            mN = 1.0

        method = 0

        wavenumber = numpy.pi*2/wavefront.get_wavelength()

        x = wavefront.get_abscissas()


        if mN != 1.0:
            npoints_exit = int(mN * x.size)
        else:
            npoints_exit = x.size

        detector_abscissas = numpy.linspace(mX*x[0],mX*x[-1],npoints_exit)

        if method == 0:
            # calculate via loop pver detector coordinates
            x1 = wavefront.get_abscissas()
            x2 = detector_abscissas
            fieldComplexAmplitude = numpy.zeros_like(x2,dtype=complex)
            for ix,x in enumerate(x2):
                r = numpy.sqrt( numpy.power(x1-x,2) + numpy.power(propagation_distance,2) )
                distances_array  = numpy.exp(1.j * wavenumber *  r)
                fieldComplexAmplitude[ix] = (wavefront.get_complex_amplitude() * distances_array).sum()
        elif method == 1:
            # calculate via outer product, it spreads over a lot of memory, but it is OK for 1D
            x1 = numpy.outer(wavefront.get_abscissas(),numpy.ones(detector_abscissas.size))
            x2 = numpy.outer(numpy.ones(wavefront.size()),detector_abscissas)
            r = numpy.sqrt( numpy.power(x1-x2,2) + numpy.power(propagation_distance,2) )

            distances_matrix  = numpy.exp(1.j * wavenumber *  r)
            fieldComplexAmplitude = numpy.dot(wavefront.get_complex_amplitude(),distances_matrix)


        wavefront_out =  GenericWavefront1D(wavefront.get_wavelength(), ScaledArray.initialize_from_steps(fieldComplexAmplitude,
                                                                                                detector_abscissas[0],
                                                                                                detector_abscissas[1]-detector_abscissas[0]))

        # added srio@esrf.eu 2018-03-23 to conserve energy - TODO: review method!
        wavefront_out.rescale_amplitude( numpy.sqrt(wavefront.get_intensity().sum() /
                                                    wavefront_out.get_intensity().sum()))

        return wavefront_out




import numpy

from wofry.propagator.propagator import PropagationManager, PropagationElements, PropagationParameters
from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D
from wofry.propagator.propagators2D.fresnel import Fresnel2D
from wofry.propagator.propagators2D.fresnel_convolution import FresnelConvolution2D
from wofry.propagator.propagators2D.fraunhofer import Fraunhofer2D
from wofry.propagator.propagators2D.integral import Integral2D
from wofry.propagator.propagators2D.fresnel_zoom_xy import FresnelZoomXY2D
from wofry.beamline.optical_elements.ideal_elements.screen import WOScreen
from syned.beamline.element_coordinates import ElementCoordinates
from syned.beamline.beamline_element import BeamlineElement

from srxraylib.plot.gol import plot_image, plot

def initialize_default_propagator_2D():
    propagator = PropagationManager.Instance()

    propagator.add_propagator(Fraunhofer2D())
    propagator.add_propagator(Fresnel2D())
    propagator.add_propagator(FresnelConvolution2D())
    propagator.add_propagator(Integral2D())
    propagator.add_propagator(FresnelZoomXY2D())

def guess_wavefront_curvature(wavefront,radius = 100.0):

    complex_amplitude = wavefront.get_complex_amplitude()


    new_complex_amplitude = complex_amplitude * numpy.exp( -1.0j * wavefront.get_wavenumber() *
                            (wavefront.get_mesh_x()**2 + wavefront.get_mesh_y()**2) / (-2*radius) )



if __name__ == "__main__":

    try:
        initialize_default_propagator_2D()
    except:
        pass

    propagator = PropagationManager.Instance()


    wavefront = GenericWavefront2D.initialize_wavefront_from_range(x_min=-2.5e-3,
                                                                   x_max=2.5e-3,
                                                                   y_min=-1e-3,
                                                                   y_max=1e-3,
                                                                   number_of_points=(2*1024, 1024),
                                                                   wavelength=73e-12)

    radius = 28.3

    wavefront.set_spherical_wave(radius=radius)

    scale_factor = 1

    screen = WOScreen(name="PIRRONE")
    coordinates = ElementCoordinates(p = 0.0, q=-scale_factor*radius)

    propagation_elements = PropagationElements()
    propagation_elements.add_beamline_element(BeamlineElement(optical_element=screen,
                                                              coordinates=coordinates))

    parameters = PropagationParameters(wavefront=wavefront,
                                       propagation_elements=propagation_elements)
    parameters.set_additional_parameters("shift_half_pixel", 1)
    parameters.set_additional_parameters("magnification_x", 0.005)
    parameters.set_additional_parameters("magnification_y", 0.005)

    output_wf_1 = propagator.do_propagation(propagation_parameters=parameters,
                                            handler_name=Fresnel2D.HANDLER_NAME)

    output_wf_2 = propagator.do_propagation(propagation_parameters=parameters,
                                            handler_name=FresnelZoomXY2D.HANDLER_NAME)

    plot_image(output_wf_1.get_intensity(), output_wf_1.get_coordinate_x()*1e6 , output_wf_1.get_coordinate_y()*1e6, title="WF1 - "+Fresnel2D.HANDLER_NAME)
    plot_image(output_wf_2.get_intensity(), output_wf_2.get_coordinate_x()*1e6 , output_wf_2.get_coordinate_y()*1e6, title="WF2 - "+FresnelZoomXY2D.HANDLER_NAME)





import numpy

from syned.beamline.optical_elements.absorbers.slit import Slit
from syned.beamline.shape import BoundaryShape, Rectangle, Circle, Ellipse

from wofry.beamline.decorators import OpticalElementDecorator

class WOSlit(Slit, OpticalElementDecorator):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        Slit.__init__(self, name=name, boundary_shape=boundary_shape)

    def applyOpticalElement(self, wavefront, parameters=None):
        boundaries = self._boundary_shape.get_boundaries()

        if isinstance(self._boundary_shape, Rectangle):
            wavefront.clip_square(boundaries[0], boundaries[1], boundaries[2], boundaries[3])
        elif isinstance(self._boundary_shape, Circle):
            wavefront.clip_circle(boundaries[0], boundaries[1], boundaries[2])
        else:
            raise NotImplementedError("to be implemented")

        return wavefront

class WOGaussianSlit(Slit, OpticalElementDecorator):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        Slit.__init__(self, name=name, boundary_shape=boundary_shape)

    def applyOpticalElement(self, wavefront, parameters=None):
        boundaries = self._boundary_shape.get_boundaries()
        aperture_diameter_x =  numpy.abs(boundaries[1] - boundaries[0])
        aperture_diameter_y =  numpy.abs(boundaries[2] - boundaries[3])
        X = wavefront.get_mesh_x()
        Y = wavefront.get_mesh_y()

        wavefront.rescale_amplitudes(numpy.exp(- (((X*X)/2/(aperture_diameter_x/2.35)**2) + \
                                                  ((Y*Y)/2/(aperture_diameter_y/2.35)**2) )))

        return wavefront

class WOSlit1D(Slit, OpticalElementDecorator):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        Slit.__init__(self, name=name, boundary_shape=boundary_shape)

    def applyOpticalElement(self, wavefront, parameters=None):
        boundaries = self._boundary_shape.get_boundaries()

        if isinstance(self._boundary_shape, Rectangle):
            wavefront.clip(boundaries[0], boundaries[1])
        else:
            raise NotImplementedError("to be implemented")

        return wavefront

class WOGaussianSlit1D(Slit, OpticalElementDecorator):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        Slit.__init__(self, name=name, boundary_shape=boundary_shape)

    def applyOpticalElement(self, wavefront, parameters=None):
        boundaries = self._boundary_shape.get_boundaries()
        aperture_diameter =  numpy.abs(boundaries[1] - boundaries[0])
        X = wavefront.get_abscissas()

        window = numpy.exp(-(X*X)/2/(aperture_diameter/2.35)**2)
        wavefront.rescale_amplitudes(window)

        return wavefront
